-- Henge Action handlers
local AceOO = AceLibrary("AceOO-2.0")
local Dewdrop = AceLibrary("Dewdrop-2.0")
local PT = AceLibrary("PeriodicTable-2.0")

local actionTable = {}

HengeAction = AceOO.Class()
HengeAction.virtual = true

local HengeActionNone = AceOO.Class(HengeAction)
local HengeActionSpell = AceOO.Class(HengeAction)
local HengeActionItem = AceOO.Class(HengeAction)
local HengeActionOption = AceOO.Class(HengeAction)
local HengeActionFunc = AceOO.Class(HengeAction)

		--------------------------
		-- Data structure notes --
		--------------------------
--[[ HengeAction definition tables
	Must contain the following:
		'id' - a string the uniquely identifies the action
		'type' - a string giving the type of action, as used in Blizzard's
			SecureActionButton API, or 'none'.
		One of 'spell', 'item', or 'menu' - depends on the value of 'type',
			again used by Blizzard's SecureActionButton API.  Not needed if
			type is 'none'.
]]--

		---------------------------------
		-- HengeAction local functions --
		---------------------------------
local SN = function(k)
	return Henge:SpellName(k)
end

local function ConvertClick(mbs)
	local prefix = ""
	local suffix = ""
	for k in string.gmatch(mbs, "(%w+)") do
		if k == "Ctrl" then
			prefix = prefix.."ctrl-"
		elseif k == "Alt" then
			prefix = prefix.."alt-"
		elseif k == "Shift" then
			prefix = prefix.."shift-"
		elseif k == "LeftButton" then
			suffix = "1"
		elseif k == "RightButton" then
			suffix = "2"
		elseif k == "MiddleButton" then
			suffix = "3"
		end
	end
	return prefix, suffix
end

local function Attr(key, mbs)
	if type(key) ~= "string" or type(mbs) ~= "string" then
		Henge:Debug("Attr() - Invalid argument(s).")
		return nil
	end
	local prefix, suffix = ConvertClick(mbs)
	local attr = nil
	if prefix then
		attr = prefix
	else
		attr = ""
	end
	attr = attr..key
	if suffix then
		attr = attr..suffix
	end
	return attr
end

		--------------------------------
		-- HengeAction main functions --
		--------------------------------

--[[ HengeAction:Add(arg) - Adds the action defined by 'arg' to the available
	actions for buttons/spheres/etc. to use.
		'arg' - Must be a _HengeAction definition table_
]]--
function HengeAction:Add(arg)
	if type(arg) ~= "table" or
	type(arg.id) ~= "string" or
	type(arg.type) ~= "string" then
		Henge:Debug("%s:Add() - Invalid argument.", self)
		return
	end
	if type(actionTable[arg.id]) == "table" then
		Henge:Debug("%s:Add() - Action '%s' already exists.", self, arg.id)
		return
	end
	if arg.type == 'none' then
		actionTable[arg.id] = HengeActionNone:new(arg)
	elseif arg.type == 'spell' then
		if type(arg.spell) ~= "string" then
			Henge:Debug("%s:Add() - Invalid argument(s) for Spell.", self)
			return
		end
		actionTable[arg.id] = HengeActionSpell:new(arg)
	elseif arg.type == 'item' then
		if type(arg.item) ~= "string" and type(arg.itemid) ~= "number" then
			Henge:Debug("%s:Add() - Invalid argument(s) for Item.", self)
			return
		end
		actionTable[arg.id] = HengeActionItem:new(arg)
	elseif arg.type == 'option' then
		if type(arg.option) ~= "string" then
			Henge:Debug("%s:Add() - Invalid argument(s) for OMenu.", self)
			return
		end
		actionTable[arg.id] = HengeActionOption:new(arg)
	else
		Henge:Debug("%s:Add() - Type '%s' not recognized.", self, arg.type)
		return
	end
end

--[[ HengeAction:Get(id) - Returns the HengeAction specified.
		'id' - Must be the 'id' string of an available action
]]--
function HengeAction:Get(id)
	if type(id) ~= "string" then
		Henge:Debug("%s:Get() - Invalid argument.", self)
		return nil
	end
	if not AceOO.inherits(actionTable[id], HengeAction) then
		Henge:Debug("%s:Get() - Action '%s' is not valid.", self, id)
		return nil
	end
	return actionTable[id]
end

function HengeAction:Update(at)
	for k, v in pairs(actionTable) do
		if (not at) or v.type == at then
			v:Update()
		end
	end
end

function HengeAction:Clear(frame, mbs)
	self.hbobj.frame:SetAttribute(self:Attr("type"), ATTRIBUTE_NOOP)
	self.hbobj.frame:SetAttribute(self:Attr("spell"), ATTRIBUTE_NOOP)
	self.hbobj.frame:SetAttribute(self:Attr("item"), ATTRIBUTE_NOOP)
end

		---------------------------------------------------
		-- HengeAction - virtual base class for actions --
		---------------------------------------------------
function HengeAction:ToString()
	return "HengeAction"
end

function HengeAction.prototype:init(opt)
	HengeAction.super.prototype.init(self)

	for k, v in pairs(opt) do
		if not self[k] then
			self[k] = v
		else
			Henge:Debug("%s:init() - Attempt to overwrite '%s'.", self, k)
		end
	end
end

function HengeAction.prototype:Apply(frame, mbs)
	Henge:Debug(
		"%s:Apply() - Subclass for '%s' must provide this function.",
		self, self.type
	)
end

function HengeAction.prototype:ActionInfo()
	local icon = nil
	if type(self.icon) == "string" then
		icon = self.icon
	else
		icon = "Interface\\AddOns\\Henge\\icon"
	end
	return self.name, self.desc, self.icon
end

		-----------------------------------------------
		-- HengeActionNone - Placeholder action type --
		-----------------------------------------------

function HengeActionNone:ToString()
	return "HengeActionNone"
end

function HengeActionNone.prototype:init(opt)
	HengeActionNone.super.prototype.init(self, opt)
end

function HengeActionNone.prototype:Apply(frame, mbs)
end

		---------------------------------------------------------------
		-- HengeActionSpell - Handles casting spells/using abilities --
		---------------------------------------------------------------

function HengeActionSpell:ToString()
	return "HengeActionSpell"
end

function HengeActionSpell.prototype:init(opt)
	HengeActionSpell.super.prototype.init(self, opt)
	Henge:LearnSpellFunc(self.id, function() self:OnLearnSpell() end)
	self:OnLearnSpell()
end

function HengeActionSpell.prototype:OnLearnSpell()
	if self.spellID and self.icon then return end
	for tloop = 1, 180 do
		local name, _, offset, count = GetSpellTabInfo(tloop)
		if not name then
			break
		end
		for sloop = offset + 1, offset + count do
			local sn, sr = GetSpellName(sloop, BOOKTYPE_SPELL)
			if self.spell == sn and ((not self.rank) or sr == self.rank) then
				self.spellID = sloop
				self.icon = GetSpellTexture(sloop, BOOKTYPE_SPELL)
			end
		end
	end
end

function HengeActionSpell.prototype:Apply(frame, mbs)
	frame:SetAttribute(Attr("type", mbs), "spell")
	frame:SetAttribute(Attr("spell", mbs), self.spell)
end

		--------------------------------------------------------------
		-- HengeActionItem - Handles using consumable/charged items --
		--------------------------------------------------------------

function HengeActionItem:ToString()
	return "HengeActionItem"
end

function HengeActionItem.prototype:init(opt)
	HengeActionItem.super.prototype.init(self, opt)
	if type(opt.itemid) == "number" then
		local name, _, _, _, _, _, _, _, _, icon = GetItemInfo(opt.itemid)
		self.item = name
		if type(self.name) ~= "string" then
			self.name = name
		end
		if type(self.icon) ~= "string" then
			self.icon = icon
		end
	end
end

function HengeActionItem.prototype:Apply(frame, mbs)
	frame:SetAttribute(Attr("type", mbs), "item")
	frame:SetAttribute(Attr("item", mbs), self.item)
end

		--------------------------------------------------
		-- HengeActionOption - Handles user-set actions --
		--------------------------------------------------
function HengeActionOption:ToString()
	return "HengeActionOption"
end

function HengeActionOption.prototype:init(opt)
	HengeActionOption.super.prototype.init(self, opt)
	if not HengeAction.db then
		HengeAction.db = Henge:AcquireDBNamespace("Action")
	end
	if type(HengeAction.db.char[self.option]) ~= "string" then
		HengeAction.db.char[self.option] = self.default
	end
end

function HengeActionOption.prototype:Apply(frame, mbs)
	local action = HengeAction:Get(Henge.db.char.actopt[self.option])
	if not AceOO.inherits(action, HengeAction) then
		return
	end
	action:Apply(frame, mbs)
end

function HengeActionOption.prototype:ActionInfo()
	if type(Henge.db.char.actopt[self.option]) ~= "string" then
		Henge.db.char.actopt[self.option] = self.default
	end
	local action = HengeAction:Get(Henge.db.char.actopt[self.option])
	if not AceOO.inherits(action, HengeAction) then
		return nil, nil, nil
	end
	return action:ActionInfo()
end

function HengeActionOption.prototype:GetOption()
	return HengeAction.db.char[self.option]
end

function HengeActionOption.prototype:SetOption(arg)
	HengeAction.db.char[self.option] = arg
end

		-------------------------------------------------------
		-- HengeActionFunc - Handles function calls on click --
		-------------------------------------------------------
function HengeActionFunc:ToString()
	return "HengeActionFunc"
end

function HengeActionFunc.prototype:init(opt)
	HengeActionFunc.super.prototype.init(self, opt)
end

function HengeActionFunc.prototype:Apply(frame, mbs)
	frame:SetAttribute(Attr("type", mbs), "menu")
	frame.menu = self.func
end
